/*
  Tutanchamun, an arcade game
  Copyright (C) 1998  David Kastrup
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>
#include "tutan.h"
#include "tutdisp.h"
#include "tutsound.h"

#ifndef HIGHNAME
#define HIGHNAME "/var/lib/games/tutscore.dat"
#endif

FILE *highfp = NULL;

const char allchars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ. ";

struct _highscore {
  char name[3];
  char score[5];
} higall[10], higtod[10] = {
  {"MAD","15234"},
  {"BAT","13124"},
  {"AND","11388"},
  {"BAD"," 9510"},
  {"RAT"," 7846"},
  {"HIT"," 6744"},
  {"THE"," 5432"},
  {"BIG"," 4876"},
  {"FAT"," 4148"},
  {"CAT"," 3272"}
};


int scocmp(char a[5], char b[5])
{
  if (a[0] == b[0])
    return memcmp(a+1,b+1,4);
  if (a[0] == '0')
    return 1;
  if (b[0] == '0')
    return -1;
  return a[0] > b[0] ? 1 : -1;
}

void gethig()
{
  int i, j, digits, id;

  umask(022);
  /* This is stupid!  Is there really no other way to open read/write
     without truncating an existing file but creating a non-existing
     one? */
  id = open(HIGHNAME, O_CREAT|O_RDWR, 0644);
  if (id >= 0)
  highfp = fdopen(id, "r+");
  if (highfp != NULL) {
    if (fread(higall, sizeof(higall), 1, highfp) == 1) {
      for (i=0; i<10; i++) {
	for (j=0; j<3; j++) {
	  if (higall[i].name[j] == '\0'
	      || strchr(allchars, higall[i].name[j]) == NULL)
	    goto bad;
	}
	digits = 0;
	/* We have the following states of digits:
	   0 means we're still skipping blanks
	   1 means that we have caught a leading 0 not at the start of
	     the highscore (meaning only 1-9 can come after)
	   2 means that all digits may follow (the score wraparound at
	    1000000 is interpreted correctly as long as we have not
	    reached 1100000)
	 */
	for (j=0; j<5; j++) {
	  if (higall[i].score[j] == ' ' && digits == 0)
	    continue;
	  if (j != 0 && higall[i].score[j] == '0') {
	    switch(digits) {
	    case 0:
	      digits = 1;
	      continue;
	    case 1:
	      goto bad;
	    case 2:
	      continue;
	    }
	  }
	  digits = 2;
	  if (strchr("0123456789",higall[i].score[j]) == NULL)
	    goto bad;
	}
	if (digits == 0)
	  goto bad;
	if (scocmp(higall[i].score, higtod[i].score) < 0)
	  goto bad;
      }
      for (i=0; i<9; i++) {
	if (scocmp(higall[i].score, higall[i+1].score) < 0)
	  goto bad;
      }
      return;
    }
  }
bad:
  memcpy(higall,higtod,sizeof(higall));
}

int hmdflg = 0;

void puthig()
{
  if (!hmdflg || highfp == NULL)
    return;
  rewind(highfp);
  fwrite(higall, sizeof(higall), 1, highfp);
  fflush(highfp);
}


void shohig()
{
  static char num[] = " 1.ABC 123450 ABC 123450";
  int i;

  d_wrbig("HIGHSCORES", 4, 2);
  d_write("ALL  TIMES   TODAYS", 6, 8);
  d_write("GREATEST     BEST", 8, 10);

  for (i=0; i<10; i++) {
    if (i < 9) {
      num[0] = ' ';
      num[1] = '1'+i;
    } else {
      num[0] = '1';
      num[1] = '0';
    }
    memcpy(num+3,higall[i].name, 3);
    memcpy(num+7,higall[i].score, 5);
    memcpy(num+14,higtod[i].name, 3);
    memcpy(num+18,higtod[i].score, 5);
    d_write(num, 0, 3*i+14);
  }
}

void chkhig()
{
  int i, ind;
  static char buf[2];
  struct _highscore scorebuf;

  if (scocmp(playfield[player].score,higtod[9].score) <= 0)
    return;
  memcpy(scorebuf.score, playfield[player].score, 5);
  sndhig();
  playms(player);
  d_write("YOU REACHED THE CHARTS.", 0, 24);
  d_write("PLEASE ENTER YOUR NAME", 0, 27);
  d_write("BY PUSHING THE JOYSTICK", 0, 30);
  d_write("LEFT OR RIGHT TO SELECT", 0, 33);
  d_write("A LETTER AND SMART BOMB", 0, 36);
  d_write("KEY TO ENTER IT.", 0, 39);
  for (i=0; i<3; i++) {
    ind = 0;
    for (;;) {
      buf[0] = allchars[ind];
      d_wrbig(buf, 16+4*i, 44);
      music();
      if (d_smarky()) {
	scorebuf.name[i] = buf[0];
	break;
      }
      switch (d_joystc()) {
      case 1:
	if (allchars[++ind])
	  continue;
	ind = 0;
	continue;
      case 2:
	if (ind) {
	  --ind;
	  continue;
	}
	ind = sizeof(allchars)-2;
	continue;
      }
    }
  }
  clrscr();
  if (scocmp(scorebuf.score, higall[9].score) > 0) {
    hmdflg = 1;
    for (i=8; i>=0; --i) {
      if (scocmp(scorebuf.score, higall[i].score) <= 0)
	break;
    }
    if (++i < 9)
      memmove(&higall[i+1],
	      &higall[i],
	      sizeof(*higall)*(9-i));
    memcpy(&higall[i], &scorebuf, sizeof(scorebuf));
    d_drawshot(6, 16 + 3*i,20,7);
  }
  for (i=8; i>=0; --i) {
    if (scocmp(scorebuf.score, higtod[i].score) <= 0)
      break;
  }
  if (++i < 9)
    memmove(&higtod[i+1],
	    &higtod[i],
	    sizeof(*higtod)*(9-i));
  memcpy(&higtod[i], &scorebuf, sizeof(scorebuf));
  d_drawshot(28, 16 + 3*i,20,7);
  shohig();
  paus(8);
}
